;Http interface driver for lookout

(use-modules (ice-9 debug)
             
	     (srfi srfi-1)
	     (srfi srfi-2)
	     (srfi srfi-13)
	     (alterator str)
             (alterator object)
             (alterator http dialog)
             (alterator http session)
             (alterator http auth))

;;;;;;;;;;;;;;;;;;;;;;;; server engine
(define *server-socket* #f)

(define (delete-http-application app)
  (define-operation close)
  (close app))



(define (create-and-bind-socket socket-name)

  (define (close-and-throw s args)
    (close s)
    (apply throw 're-send-system-error args))

  (define (bind-and-return s)
    (bind s AF_UNIX socket-name) s)
  
  (let ((s (socket PF_UNIX SOCK_STREAM 0)))
    (catch 'system-error
      (lambda() (bind-and-return s))
      (lambda(key . args)
        (if (and (string? (car args)) (string=? (car args) "bind")
                 (number? (car (cadddr args))) (= (car (cadddr args)) 98))
            (catch 'system-error
              (lambda ()
                (connect s AF_UNIX socket-name)
                (close-and-throw s args)); oops ... something uses this socket
              (lambda(key2 . args2) ;ok remove and rebind
                (delete-file socket-name)
                (bind-and-return s)))
            (close-and-throw s args))))))

(define (make-server-socket socket-name)
    (let ((s (create-and-bind-socket socket-name)))
      (listen s 10)
      (chmod socket-name #o666) ; temporary hack, will be replaced with chown and chmod to group
      (object
       #f
       ((close self) (close s) (delete-file socket-name))
       ((get-connection self) (car (accept s))))))
    
;create new server
(define (make-http-application . args)
  (let* ((socket-name (if (null? args) "/tmp/alterator-sock" (car args))))
    (format #t "socket-name=~A~%" socket-name)
    (set! *server-socket* (make-server-socket socket-name))))

;listen on socket and run appropriate processing
(define (exec-http-application)
  (define-operation write-input)
  (define-operation read-output)
  (define-operation get-connection)
  
  (write-line "listen for connections ...")
  (let listen-loop ((c (get-connection *server-socket*)))
    (write-line "accepting connection and process commands")
    (let command-loop ((cmd (read c)))
      (and (not (eof-object? cmd))
           (begin (write (auth-http-request cmd
                                            (lambda(msg)
                                              (let ((session (fluid-ref http-session)))
                                                (write-input session msg)
                                                (read-output session))))
                         c)
                  (command-loop (read c)))))
    (write-line "go to next processing")
    (listen-loop (get-connection *server-socket*))))


;;;;;;;;;;;;;;;;;;;;;;;;;; widget engines
(define (extract-real-id x)
  (define-operation get-id)
  (if (object? x) (get-id x) x))

(define (extract-all-real-ids lst)
  (map extract-real-id lst))

;for widgets: we save widgets as a properties: send to http as an ids,or as a real ids
(define (make-http-default-widget id parend-id)
  (define-operation silent-set)
  

  (let ((attributes '())
        (apriori-attributes '((activity . (#t))
                              (visibility . (#t)))))
    (object
      #f
      ((set self name . value)
       (let ((value (extract-all-real-ids value)))
         (apply silent-set self name value)
         (http-post-event 'action "set"
                          'widget-id id
                          'attr name
                          'value (string-join (map sure-string value) ";"))))
      ((silent-set self name . value)
       (let ((value (extract-all-real-ids value)))
         (set! attributes (alist-set
                           (sure-symbol name)
                           value
                           attributes))))
      ((get self name)
       (cond ((or (assq (sure-symbol name) attributes)
                  (assq (sure-symbol name) apriori-attributes)) =>
                  (lambda (x)
                    (let ((x  (cdr x)))
                      (if (= (length x) 1)
                          (car x)
                          x))))
	     (else #f)))
      ((bubbling self subcommand . args)
       (case subcommand
	 ((re-create)
	  (for-each (lambda (attr)
		      (let ((name (car attr))
			    (value (cdr attr)))
                        (and (not (eq? name 'widget-id))
                             (http-post-event 'action "set"
                                              'widget-id id
                                              'attr name
                                              'value (string-join (map sure-string
                                                                       value) ";")))))
		    attributes)))))))


(define (notify-add-item id item)
  (http-post-event 'action "set"
                   'widget-id id
                   'attr "append-item"
                   'value (string-append (car item) ";" (cdr item))))

(define (make-http-listbox-setter id items)
  (define-operation append-item)
  (define-operation remove)
  (define-operation set)

  (object
   #f
   ((items self itemlist)
    (remove self 'all)
    (map (lambda (item) (append-item self item)) itemlist))
   ((append-item self item)
    (cell-set! items (append (cell-ref items) (list (if (pair? item)
    							;made copies to avoid conflicts in future changes
							;throw set-car! and set-cdr!
                                                        (cons (car item) (cdr item))
                                                        (cons item "")))))
    (notify-add-item id (car (last-pair (cell-ref items)))))
   ((remove self . range) (if (eq? (car range) 'all)
                               (begin
                                 (cell-set! items '())
                                 (http-post-event 'action "set"
                                                  'widget-id id
                                                  'attr "remove-item"
                                                  'value "all"))
                               (for-each (lambda(number)
                                           (cell-set! items (list-remove (cell-ref items) number))
                                           (http-post-event 'action "set"
                                                            'widget-id id
                                                            'attr "remove-item"
                                                            'value (sure-string number)))
                                         (reverse (extract-range range)))))
   ((item-pixmap self pixmap . range) (for-each (lambda(number)
                                                  (set-cdr! (list-ref (cell-ref items) number) pixmap)
                                                  (http-post-event 'action "set"
                                                                   'widget-id id
                                                                   'attr "item-pixmap"
                                                                   'value (string-append
								           pixmap
									   ";"
								           (sure-string number))))
                                                (extract-range range)))
   ((item-text self text . range) (for-each (lambda (number)
                                              (set-car! (list-ref (cell-ref items) number) text)
                                              (http-post-event 'action "set"
                                                               'widget-id id
                                                               'attr "item-text"
                                                               'value (string-append
							               text
								       ";"
							               (sure-string number))))
                                            (extract-range range)))))

(define (make-http-listbox-getter id items default-widget)
  (define-operation get)
  (define-operation item-text)
  (object
   #f
   ((count self) (length (cell-ref items)))
   ((text self) (item-text self (get default-widget 'current)))
   ((item-text self number) (car (list-ref (cell-ref items) number)))
   ((item-pixmap self number) (cdr (list-ref (cell-ref items) number)))))

(define (make-http-listbox id parent-id)
  (define-operation set)
  (define-operation get)
  (define-operation bubbling)
  
  (let* ((items (make-cell '()))
         (default-widget (make-http-default-widget id parent-id))
         (http-listbox-setter (make-http-listbox-setter id items))
         (http-listbox-getter (make-http-listbox-getter id items default-widget)))
    (join
     (object
      #f
      ((set self value . args)
       (let ((op (select-operation http-listbox-setter value)))
         (if (procedure? op)
             (apply op http-listbox-setter args)
             (apply set default-widget value args))))
      ((get self value . args)
       (let ((op (select-operation http-listbox-getter value)))
         (if (procedure? op)
             (apply op http-listbox-getter args)
             (apply get default-widget value args))))
      ((bubbling self subcommand . args)
       (case subcommand
         ((re-create)
          (http-post-event 'action "set"
                           'widget-id id
                           'attr "remove-item"
                           'value "all")
          (for-each (lambda (item) (notify-add-item id item)) (cell-ref items))))
       (apply bubbling default-widget subcommand args)))
      default-widget)))

(define (make-http-widget type . rest)
  (define-operation get-widget)
  (define-operation add-widget)
  (define-operation remove-widget)
  (define-operation make-widget-id)
  (define-operation list-events)
  (define-operation bubbling)
  
  (lambda (eventholder parent-widget)

    (define (register-events id)
      (for-each (lambda (x)
		  (http-post-event 'action "create-event"
				   'widget-id id
				   'value x))
		(list-events eventholder)))

    (let* ((id (make-widget-id (fluid-ref http-session)))
	  (parent-id (or (and parent-widget (get-widget parent-widget)) ""))
	  (mixed-widget  (if (null? rest) (make-http-default-widget id parent-id) ((car rest) id parent-id))))

      (http-post-event 'action "new" 'widget-id id 'type type 'parent parent-id)
      
      (join
	(object
	  #f
	  ((get-widget self) id)
	  ((post-new self super . args)
           (add-widget (fluid-ref http-session) id super)
	   (register-events id))
	  ((delete self)
           (remove-widget (fluid-ref http-session) id)
           (http-post-event 'action "delete" 'widget-id id))
	  ((bubbling self subcommand . args)
	   (case subcommand
	     ((re-create)
	      (and id
		   (begin
		     (http-post-event 'action "new"
				      'widget-id id
				      'type type
				      'parent parent-id)
		     (bubbling mixed-widget subcommand)
		     (register-events id)))))))
	mixed-widget))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;driver description
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declare-widget button (make-http-widget "button") text)
(declare-widget label (make-http-widget "label") text)
(declare-widget edit (make-http-widget "edit") text)
(declare-widget hbox (make-http-widget "hbox"))
(declare-widget vbox (make-http-widget "vbox"))
(declare-widget groupbox (make-http-widget "groupbox") title)
(declare-widget textbox (make-http-widget "textbox") text)
(declare-widget tab-page (make-http-widget "tab-page") text)
(declare-widget tabbox (make-http-widget "tabbox"))
(declare-widget checkbox (make-http-widget "checkbox") text)

;wrapper over tab-page
(define-macro (tab <name> . <args>)
   `(tab-page ,<name> (vbox . ,<args>)))

(declare-widget listbox (make-http-widget "listbox" make-http-listbox))

(declare-widget combobox (make-http-widget "combobox" make-http-listbox))

(declare-application-constructor make-http-application)
(declare-application-destructor delete-http-application)
(declare-application-runner exec-http-application)

(declare-root-widget (make-http-widget "vbox"))
(declare-popup-widget (make-http-widget "dialog" (lambda (id parent-id)
                                                   (make-http-dialog
                                                    (make-http-default-widget id parent-id)
                                                    id
                                                    parent-id))))

(declare-lang-selector (lambda()
                         (define-operation get-lang)
                         (list (or (get-lang (fluid-ref http-session)) "en"))))
